(*
    A proof assistant inside Coq for propositional logic
    Copyright (C) 2024  Max Ole Elliger

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*)
Require Import Bool String List Classical.

Inductive form := 
  | f_bot : form
  | f_atom : string -> form
  | f_neg : form -> form
  | f_conj : form -> form -> form.

Fixpoint f_satisfaction (k : string -> bool) (f : form) : Prop :=
  match f with
  | f_bot => False
  | f_atom a => if k a then True else False
  | f_neg f1 => ~ (f_satisfaction k f1)
  | f_conj f1 f2 => (f_satisfaction k f1) /\ (f_satisfaction k f2)
  end.

Definition f_valid (f : form) : Prop :=
  forall k,
    f_satisfaction k f.

Definition f_satisfiable (f : form) : Prop :=
  exists k,
    f_satisfaction k f.

Fact f_valid_satisfiable :
  forall f,
    f_valid f <-> ~ (f_satisfiable (f_neg f)).
Proof.
  unfold f_valid,f_satisfiable.
  intros f.
  split.
  -
    intros H1 [k H2].
    apply H2.
    fold f_satisfaction.
    exact (H1 k).
  -
    intros H1.
    Search (~ (exists _, _)).
    apply not_ex_not_all.
    intros [k H2].
    apply H1.
    exists k.
    exact H2.
Qed.

Definition f_conseq (fs : list form) (f : form) : Prop :=
  forall k,
    (forall f, In f fs -> f_satisfaction k f) ->
    f_satisfaction k f.

Fact f_valid_conseq : 
  forall f,
    f_valid f <-> f_conseq nil f.
Proof.
  firstorder.
Qed.

Fact f_neg_conseq_satisfiable :
  forall ls f,
  f_conseq ls (f_neg f) <-> f_conseq (f :: ls) f_bot.
Proof.
  intros ls f.
  split.
  -
    firstorder.
  -
    intros H1 k H2 H3.
    eapply H1.
    firstorder.
    subst.
    assumption.
Qed.

Inductive proof : list form -> form -> Set :=
  | reit :
      forall ls f1,
        In f1 ls ->
        proof ls f1
  | botI :
      forall ls f1,
        proof ls f1 ->
        proof ls (f_neg f1) ->
        proof ls f_bot
  | botE :
      forall ls f1,
        proof ls f_bot ->
        proof ls f1
  | negI :
      forall ls f1,
        proof (f1 :: ls) f_bot ->
        proof ls (f_neg f1)
  | negE :
      forall ls f1,
        proof ls (f_neg (f_neg f1)) ->
        proof ls f1
  | conjI :
      forall ls f1 f2,
        proof ls f1 ->
        proof ls f2 ->
        proof ls (f_conj f1 f2)
  | conjE1 : 
      forall ls f1 f2,
        proof ls (f_conj f1 f2) ->
        proof ls f1
  | conjE2 :
      forall ls f1 f2,
        proof ls (f_conj f1 f2) ->
        proof ls f2.

(*
Ltac reit n :=
  apply reit;
  do n right;
  left; reflexivity.
 *)

Theorem soundness : 
  forall ls f,
    proof ls f ->
    f_conseq ls f.
Proof.
  induction 1 as
  [ls f1 H1
  |ls f1 H1 IH1 H2 IH2
  |ls f1 H1 IH1
  |ls f1 H1 IH1
  |ls f1 H1 IH1
  |ls f1 f2 H1 IH1 H2 IH2
  |ls f1 f2 H1 IH1
  |ls f1 f2 H1 IH1].
  all: unfold f_conseq.
  all: firstorder.
  -
    intros H3.
    apply f_neg_conseq_satisfiable in IH1.
    eapply IH1.
    +
      eassumption.
    +
      fold f_satisfaction.
      assumption.
  -
    apply NNPP.
    eapply IH1.
    assumption.
Qed.

Print Assumptions soundness.

Example fitch_test_1 : proof nil (f_neg (f_conj (f_neg (f_atom "p")) (f_atom "p"))).
Proof.
  apply negI.
  eapply botI.
  -
    eapply conjE2.
    apply reit.
    left; reflexivity.
  -
    eapply conjE1.
    apply reit.
    left; reflexivity.
Qed.
